/*------------------------------------------------------.
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 `------------------------------------------------------+------------
 An X-based browser for color selection, based on the "forms" library
 and borrowing some code from its example snippet "colbrowser.c".
 --------------------------------------------------------------------
 $Id: ce.c,v 1.1 2004/02/04 06:59:08 loreti Exp $
 -------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include <forms.h>

#include "ce.h"

/**
 | Local prototypes (alphabetical order)
**/

void     abort(void);
void     browserSel(FL_OBJECT *, long);
void     buttonHit(FL_OBJECT *, long);
FL_FORM *createForm(void);
int      loadBrowser(const char *);
void     modColor(FL_OBJECT *, long);
int      searchEntry(void);
void     setEntry(const size_t);

/**
 | Global variables:
 | - pointers to various gadget objects;
 | - the current values of the red, green and blue color components;
 | - the text to be shown to print these components;
 | - the name of the X11 RGB database file;
 | - an array to hold the data read from this database, and its
 |   current size.
**/

FL_FORM   *pFF;                 /* Form                         */
FL_OBJECT *pRV, *pGV, *pBV;     /* Red/Green/Blue values (text) */
FL_OBJECT *pRS, *pGS, *pBS;     /* Red/Green/Blue sliders       */
FL_OBJECT *pCB;                 /* Color Box                    */
FL_OBJECT *pDBname;             /* Database name                */
FL_OBJECT *pDBbrowser;          /* Database browser             */

int     rr,  gg,  bb;           /* R/G/B integer values, 0...255    */
double rrr, ggg, bbb;           /* R/G/B double values, 0.0...255.0 */

char rv_l[] = "Red:   0.00 =   0 = 0x00"; /* R/G/B values shown */
char gv_l[] = "Green: 0.00 =   0 = 0x00";
char bv_l[] = "Blue:  0.00 =   0 = 0x00";

char db_name[FILENAME_MAX] = "/usr/X11R6/lib/X11/rgb.txt";
const char none[] = "None";

RGBdb rgbdb[MAX_RGB];
size_t rgbdbLen;

/****************************************************************************/

int main(
  int   argc,
  char *argv[]
){
  fl_initialize(&argc, argv, 0, 0, 0);
  pFF = createForm();
  fl_show_form(pFF, 0, FL_FULLBORDER, "Color Explorer");
  fl_do_forms();

  exit(EXIT_SUCCESS);
}

/****************************************************************************/

void abort(void)
{
  fl_hide_form(pFF);
  exit(EXIT_FAILURE);
}

/****************************************************************************/

void browserSel(
  FL_OBJECT *pO,
  long       arg
) {

/**
 | Call-back function for browser-related actions:
 | - a new database name (arg == DBN_CB): we ask for a new file name;
 |   try to load the browser with the file content; and finally change
 |   the file name on the main window.
 | - a selection on the database browser (arg == DBB_CB): we update
 |   accordingly the displayed color.
**/

  int         i;
  const char *pc;

  switch (arg) {
    case DBN_CB:
      pc = fl_show_input("Enter New Database Name", db_name);

      if (pc != 0   &&   strcmp(pc, db_name) != 0) {
        char buffer[FILENAME_MAX];

        strcpy(buffer, pc);
        if (loadBrowser(buffer)) {
          strcpy(db_name, buffer);
          fl_set_object_label(pDBname, db_name);
        }
      }
      break;

    case DBB_CB:
      i = fl_get_browser(pDBbrowser);
      if (i-- > 0) setEntry(i);
      break;

    default:
      sprintf(db_name, "0x%lX", arg);
      fl_show_alert("Internal error", "Invalid browser call-back code",
                    db_name, 0);
      abort();
  }
}

/****************************************************************************/

void buttonHit(
  FL_OBJECT *pFO,
  long       arg
){

/**
 | Call-back function called when a button is hit, with the
 | exception of the color-selection buttons.
 |
 | There are currently two such gadgets:
 | - the "Exit" button (arg == EB_CB), the program is terminated;
 | - the "Find nearer" button (arg == FN_CB): we search the database
 |   for the nearer color, change the browser top line if needed, and
 |   finally highlight on the browser the resulting entry.
**/

  int i;

  switch (arg) {
    case EB_CB:
      fl_hide_form(pFF);
      exit(EXIT_SUCCESS);

    case FN_CB:
      rr = fl_get_slider_value(pRS);
      gg = fl_get_slider_value(pGS);
      bb = fl_get_slider_value(pBS);

      if ((i = searchEntry()) >= 0) {
        int topLine = fl_get_browser_topline(pDBbrowser);

        if (++i < topLine   ||   i >= topLine + BR_LINES) {
          fl_set_browser_topline(pDBbrowser, i - (BR_LINES/2));
        }
        fl_select_browser_line(pDBbrowser, i);
      }
      break;

    default:
      sprintf(db_name, "0x%lX", arg);
      fl_show_alert("Internal error", "Invalid button call-back code",
                    db_name, 0);
      abort();
  }
}

/****************************************************************************/

FL_FORM *createForm(void)
{

/**
 | Calls to the library function to define the X-11 program window
**/

  FL_OBJECT *obj;

/* Form and main box */

  pFF = fl_bgn_form(FL_NO_BOX, FF_W, FF_H);
  (void) fl_add_box(FL_UP_BOX, MB_X, MB_Y, MB_W, MB_H, MB_L);

/* Red buttons, slider and values */

  obj = fl_add_button(FL_NORMAL_BUTTON,
                      R_LL_X, R_LL_Y, R_LL_W, R_LL_H, R_LL_L);
  fl_set_object_callback(obj, modColor, R_LL_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      R_L_X, R_L_Y, R_L_W, R_L_H, R_L_L);
  fl_set_object_callback(obj, modColor, R_L_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      R_R_X, R_R_Y, R_R_W, R_R_H, R_R_L);
  fl_set_object_callback(obj, modColor, R_R_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      R_RR_X, R_RR_Y, R_RR_W, R_RR_H, R_RR_L);
  fl_set_object_callback(obj, modColor, R_RR_CB);

  pRS = fl_add_slider(FL_HOR_FILL_SLIDER,
                      R_S_X, R_S_Y, R_S_W, R_S_H, R_S_L);
  fl_set_object_color(pRS, FL_COL1, FL_RED);
  fl_set_object_callback(pRS, modColor, R_S_CB);
  fl_set_slider_value(pRS, rrr);
  fl_set_slider_bounds(pRS, 0.0, 255.0);
  fl_set_slider_step(pRS, 1.0);

  pRV = fl_add_box(FL_FLAT_BOX, RV_X, RV_Y, RV_W, RV_H, rv_l);
  fl_set_object_lsize(pRV, FL_NORMAL_SIZE);
  fl_set_object_lalign(pRV, (FL_ALIGN_LEFT | FL_ALIGN_INSIDE));
  fl_set_object_lstyle(pRV, FL_FIXEDBOLD_STYLE);

/* Same for green */

  obj = fl_add_button(FL_NORMAL_BUTTON,
                      G_LL_X, G_LL_Y, G_LL_W, G_LL_H, G_LL_L);
  fl_set_object_callback(obj, modColor, G_LL_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      G_L_X, G_L_Y, G_L_W, G_L_H, G_L_L);
  fl_set_object_callback(obj, modColor, G_L_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      G_R_X, G_R_Y, G_R_W, G_R_H, G_R_L);
  fl_set_object_callback(obj, modColor, G_R_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      G_RR_X, G_RR_Y, G_RR_W, G_RR_H, G_RR_L);
  fl_set_object_callback(obj, modColor, G_RR_CB);

  pGS = fl_add_slider(FL_HOR_FILL_SLIDER,
                      G_S_X, G_S_Y, G_S_W, G_S_H, G_S_L);
  fl_set_object_color(pGS, FL_COL1, FL_GREEN);
  fl_set_object_callback(pGS, modColor, G_S_CB);
  fl_set_slider_value(pGS, ggg);
  fl_set_slider_bounds(pGS, 0.0, 255.0);
  fl_set_slider_step(pGS, 1.0);

  pGV = fl_add_box(FL_FLAT_BOX, GV_X, GV_Y, GV_W, GV_H, gv_l);
  fl_set_object_lsize(pGV, FL_NORMAL_SIZE);
  fl_set_object_lalign(pGV, (FL_ALIGN_LEFT | FL_ALIGN_INSIDE));
  fl_set_object_lstyle(pGV, FL_FIXEDBOLD_STYLE);

/* Same for blue */

  obj = fl_add_button(FL_NORMAL_BUTTON,
                      B_LL_X, B_LL_Y, B_LL_W, B_LL_H, B_LL_L);
  fl_set_object_callback(obj, modColor, B_LL_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      B_L_X, B_L_Y, B_L_W, B_L_H, B_L_L);
  fl_set_object_callback(obj, modColor, B_L_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      B_R_X, B_R_Y, B_R_W, B_R_H, B_R_L);
  fl_set_object_callback(obj, modColor, B_R_CB);
  obj = fl_add_button(FL_NORMAL_BUTTON,
                      B_RR_X, B_RR_Y, B_RR_W, B_RR_H, B_RR_L);
  fl_set_object_callback(obj, modColor, B_RR_CB);
  pBS = fl_add_slider(FL_HOR_FILL_SLIDER,
                      B_S_X, B_S_Y, B_S_W, B_S_H, B_S_L);
  fl_set_object_color(pBS, FL_COL1, FL_BLUE);
  fl_set_object_callback(pBS, modColor, B_S_CB);
  fl_set_slider_value(pBS, bbb);
  fl_set_slider_bounds(pBS, 0.0, 255.0);
  fl_set_slider_step(pBS, 1.0);

  pBV = fl_add_box(FL_FLAT_BOX, BV_X, BV_Y, BV_W, BV_H, bv_l);
  fl_set_object_lsize(pBV, FL_NORMAL_SIZE);
  fl_set_object_lalign(pBV, (FL_ALIGN_LEFT | FL_ALIGN_INSIDE));
  fl_set_object_lstyle(pBV, FL_FIXEDBOLD_STYLE);

/* Color box */

  pCB = fl_add_box(FL_DOWN_BOX, CB_X, CB_Y, CB_W, CB_H, CB_L);
  fl_mapcolor(FL_FREE_COL1, rr, gg, bb);
  fl_set_object_color(pCB, FL_FREE_COL1, FL_FREE_COL1);
  fl_set_object_dblbuffer(pCB, 1);

/* "Find nearer" button */

  obj = fl_add_button(FL_NORMAL_BUTTON, FN_X, FN_Y, FN_W, FN_H, FN_L);
  fl_set_object_lsize(obj, FL_NORMAL_SIZE);
  fl_set_object_callback(obj, buttonHit, FN_CB);

/* "Exit" button */

  obj = fl_add_button(FL_NORMAL_BUTTON, EB_X, EB_Y, EB_W, EB_H, EB_L);
  fl_set_object_lsize(obj, FL_MEDIUM_SIZE);
  fl_set_object_callback(obj, buttonHit, EB_CB);

/* Database name and browser */

  pDBbrowser = fl_add_browser(FL_HOLD_BROWSER, DBB_X, DBB_Y, DBB_W, DBB_H, "");
  fl_set_browser_fontstyle(pDBbrowser,  FL_FIXED_STYLE);
  fl_set_object_callback(pDBbrowser, browserSel, DBB_CB);

  pDBname = fl_add_button(FL_NORMAL_BUTTON, DBN_X, DBN_Y, DBN_W, DBN_H, "");
  fl_set_object_boxtype(pDBname, FL_BORDER_BOX);
  fl_set_object_callback(pDBname, browserSel, DBN_CB);
  if (loadBrowser(db_name)) {
    fl_set_object_label(pDBname, db_name);
  } else {
    fl_set_object_label(pDBname, none);
  }

  fl_end_form();
  return pFF;
}

/****************************************************************************/

int loadBrowser(
  const char *fName
) {

/**
 | Procedure to open the X11 database file and to read its data,
 | filling the database browser and the storage array.  We assume
 | that the file contains valid data; the browser names contain
 | no whitespace, and consecutive entries for the same color are
 | skipped.
**/

  FILE *fp;
  char  buffer[MAX_DB_LEN];
  int   rLast = -1, gLast = -1, bLast = -1;

  if ((fp = fopen(fName, "r")) == 0) {
    fl_show_alert("Load", fName, "Cannot be opened", 0);
    return ERROR;
  }

  rgbdbLen = 0;
  fl_freeze_form(pFF);

  while (fgets(buffer, MAX_DB_LEN, fp)) {
    int   n;
    char  name[MAX_DB_LEN];
    char *from, *to;

    if (ferror(fp)) break;
    if (buffer[0] == '!') continue;
    if (sscanf(buffer, " %d %d %d %n", &rr, &gg, &bb, &n) < 3) break;

    for (from = buffer + n,  to = name;   *from;   from++) {
      if (!isspace((unsigned char) *from)) *to++ = *from;
    }
    *to = '\0';

    if (rr == rLast   &&   gg == gLast   &&   bb == bLast) continue;

    rgbdb[rgbdbLen].r = rLast = rr;
    rgbdb[rgbdbLen].g = gLast = gg;
    rgbdb[rgbdbLen].b = bLast = bb;

    sprintf(buffer, "(%3d %3d %3d) %s", rr, gg, bb, name);
    fl_addto_browser(pDBbrowser, buffer);
    if (++rgbdbLen == MAX_RGB) break;
  }

  fclose(fp);

  if (rgbdbLen) {
    fl_set_browser_topline(pDBbrowser, 1);
    fl_select_browser_line(pDBbrowser, 1);
    setEntry(0);
    fl_unfreeze_form(pFF);
    return OK;
  } else {
    fl_unfreeze_form(pFF);
    fl_show_alert("Load", fName, "Has an invalid content", 0);
    return ERROR;
  }
}

/****************************************************************************/

void modColor(
  FL_OBJECT *pFO,
  long       arg
){

/**
 | A color gadget has been hit: a button or a slider.  Set up pointers
 | to the relevant slider and text field.
**/

  int        obj;               /* Hit family: red, blue or green  */
  int        type;              /* Hit object: LL, L, S, R or RR   */
  int        ival;              /* Slider value (integer)          */
  int       *pIval;             /* Will point to the integer value */
  double     value;             /* Slider value (new)              */
  double     old_value;         /* Slider value (old)              */
  double    *pDval;             /* Will point to the double value  */
  char      *pc;                /* Will point to the char[] value  */
  FL_OBJECT *pFO1;              /* Will point to the text values   */
  FL_OBJECT *pFO2;              /* Will point to the slider        */

  switch (obj = arg & RGB_MASK) {
    case RED:
      pFO1  = pRV;
      pFO2  = pRS;
      pIval = &rr;
      pDval = &rrr;
      pc    = rv_l;
      break;

    case GREEN:
      pFO1  = pGV;
      pFO2  = pGS;
      pIval = &gg;
      pDval = &ggg;
      pc    = gv_l;
      break;

    case BLUE:
      pFO1  = pBV;
      pFO2  = pBS;
      pIval = &bb;
      pDval = &bbb;
      pc    = bv_l;
      break;

    default:
      sprintf(db_name, "0x%X", obj);
      fl_show_alert("Internal error", "Invalid color gadget object code",
                    db_name, 0);
      abort();
  }

  old_value = value = fl_get_slider_value(pFO2);

/**
 | Update the modified value, and show changes
**/

  switch (type = (arg & LSR_MASK)) {
    case LL:
      value -= 16.0;
      if (value < 0.0) value = 0.0;
      break;

    case L:
      value -= 1.0;
      if (value < 0.0) value = 0.0;
      break;

    case R:
      value += 1.0;
      if (value > 255.0) value = 255.0;
      break;

    case RR:
      value += 16.0;
      if (value > 255.0) value = 255.0;
      break;

    case S:
      break;

    default:
      sprintf(db_name, "0x%X", type);
      fl_show_alert("Internal error", "Invalid color gadget type code",
                    db_name, 0);
      abort();
  }

  if (value != old_value) {
    fl_set_slider_value(pFO2, value);
  }

  ival = value + 0.5;
  *pIval = ival;
  *pDval = value / 255.0;

  sprintf(pc+7, "%4.2f = %3d = 0x%02X", *pDval, ival, ival);
  fl_set_object_label(pFO1, pc);

  fl_mapcolor(FL_FREE_COL1, rr, gg, bb);
  fl_redraw_object(pCB);
}

/****************************************************************************/

int searchEntry(void)
{

/**
 | Finds the database entry with the minimum "distance" from the
 | currently selected color (in rr, gg and bb).
**/

  int          i;
  int          retVal = -1;
  unsigned int minDiff = UINT_MAX;

  for (i = 0;   i < rgbdbLen;   i++) {
    int          rd, gd, bd;
    unsigned int diff;

    rd = rgbdb[i].r - rr;
    gd = rgbdb[i].g - gg;
    bd = rgbdb[i].b - bb;
    diff = R_FACTOR * rd * rd + G_FACTOR * gd * gd + B_FACTOR * bd * bd + 0.5;
    if (diff < minDiff) {
      minDiff = diff;
      retVal  = i;
    }
  }

  return retVal;
}

/****************************************************************************/

void setEntry(
  const size_t i
) {

/**
 | Sets the current color according to the i-th element of the
 | database array.
**/

  if (rr != rgbdb[i].r   ||   gg != rgbdb[i].g  ||  bb != rgbdb[i].b) {
    fl_freeze_form(pFF);

    rr = rgbdb[i].r;
    gg = rgbdb[i].g;
    bb = rgbdb[i].b;

    rrr = rr / 255.0;
    ggg = gg / 255.0;
    bbb = bb / 255.0;

    fl_mapcolor(FL_FREE_COL1, rr, gg, bb);
    fl_set_slider_value(pRS, rr);
    fl_set_slider_value(pGS, gg);
    fl_set_slider_value(pBS, bb);
    fl_redraw_object(pCB);

    sprintf(rv_l+7, "%4.2f = %3d = 0x%02X", rrr, rr, rr);
    sprintf(gv_l+7, "%4.2f = %3d = 0x%02X", ggg, gg, gg);
    sprintf(bv_l+7, "%4.2f = %3d = 0x%02X", bbb, bb, bb);
    fl_set_object_label(pRV, rv_l);
    fl_set_object_label(pGV, gv_l);
    fl_set_object_label(pBV, bv_l);

    fl_unfreeze_form(pFF);
  }
}
